from dataclasses import dataclass

import openai
from openai import OpenAI
from pydantic import BaseModel

from akaibu.paper import Paper


class IsRelevant(BaseModel):
    does_match: bool


@dataclass(frozen=True)
class Checker:
    requirement: str
    model_name: str
    base_url: str
    key: str

    def is_paper_relevant(self, paper: Paper) -> bool:
        client = OpenAI(base_url=self.base_url, api_key=self.key)
        try:
            completion = client.beta.chat.completions.parse(
                temperature=0,
                model=self.model_name,
                messages=[
                    {
                        "role": "user",
                        "content": f"""
                        Check if the following paper matches the requirements:

                        Requirements: {self.requirement}

                        Paper:
                          Title: {paper.title}
                          Abstract: {paper.abstract}
                    """,
                    }
                ],
                response_format=IsRelevant,
            )

            response = completion.choices[0].message
            if response.parsed:
                return response.parsed.does_match
            elif response.refusal:
                print(response.refusal)
            return False
        except Exception as e:
            if type(e) is openai.LengthFinishReasonError:
                print("Too many tokens: ", e)
            else:
                print(e)
            return False



import datetime
import time
from dataclasses import dataclass
from typing import Any

from feedparser import FeedParserDict
from reader import Entry


@dataclass(frozen=True)
class Paper:
    id: str
    title: str
    link: str
    abstract: str
    tags: list[dict[str, Any]]
    published: time.struct_time | datetime
    authors: str

    @classmethod
    def from_entry(cls, d: FeedParserDict | Entry) -> "Paper":
        if isinstance(d, FeedParserDict):
            return cls(
                id=d["id"],
                title=d["title"],
                link=d["link"],
                abstract=d["summary"],
                tags=d["tags"],
                published=d["published_parsed"],
                authors=d["authors"],
            )
        else:
            return cls(
                id=d.id,
                title=d.title,
                link=d.link,
                abstract=d.summary,
                tags=[],
                published=d.published,
                authors=d.author,
            )

    def to_serializable(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "title": self.title,
            "link": self.link,
            "abstract": self.abstract,
            "tags": self.tags,
            "published": str(self.published),
            "authors": self.authors,
        }


@dataclass(frozen=True)
class PaperAndSummary:
    paper: Paper
    summary: str

    def to_serializable(self) -> dict[str, Any]:
        return {
            "paper": self.paper.to_serializable(),
            "summary": self.summary,
        }

    def to_markdown(self) -> str:
        key_str = "Abstract: "
        abstract = self.paper.abstract
        abstract = abstract[abstract.find(key_str) + len(key_str) :]
        return f"""- [ ] [{self.paper.title}]({self.paper.link})
    - Summary: {self.summary}"""



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable

import numpy as np
from kensakun import Engine
from reader import Entry, Reader, make_reader

from akaibu.checker import Checker
from akaibu.paper import Paper, PaperAndSummary
from akaibu.summarizer import PaperSummarizer

os.environ["TOKENIZERS_PARALLELISM"] = "false"


@dataclass(frozen=True)
class Library:
    reader: Reader

    @classmethod
    def load_default(cls) -> "Library":
        reader = make_reader("debug.sqlite")
        return cls(reader)

    @classmethod
    def load_from_path(cls, path: Path, remake: bool = False) -> "Library":
        if remake and path.exists():
            path.unlink()
        reader = make_reader(str(path))
        return cls(reader)

    def add_url(self, url: str) -> None:
        self.reader.add_feed(url, exist_ok=True)
        self.reader.update_feeds()

    def list_urls(self) -> list[str]:
        return [feed.link for feed in self.reader.get_feeds()]

    def sort_entries(
        self, entries: Iterable[Entry], query: str, kensakun: Engine
    ) -> Iterable[Entry]:
        entries = list(entries)
        kensakun.add_documents([f"{ent.title} {ent.summary}" for ent in entries])
        _, indexes = kensakun.search(query)
        return [entries[idx] for idx in indexes]

    def get_papers(
        self,
        limit: int = 25,
        checker: Checker | None = None,
        summarizer: PaperSummarizer | None = None,
        kensakun: Engine | None = None,
    ) -> list[Paper] | list[PaperAndSummary]:
        self.reader.update_feeds()

        if kensakun:
            assert summarizer
            entries = self.reader.get_entries(limit=limit * 2, read=False)
            entries = self.sort_entries(entries, summarizer.requirement, kensakun)
        else:
            entries = self.reader.get_entries(limit=limit, read=False)

        relevant_papers = []
        for entry in entries:
            self.reader.mark_entry_as_read(entry)
            paper = Paper.from_entry(entry)

            if checker and summarizer:
                is_relevant = checker.is_paper_relevant(paper)
                if is_relevant:
                    paper_and_summary = summarizer.summarize(paper)
                    self.reader.set_tag(entry, "is_relevant", True)
                    self.reader.set_tag(
                        entry, "generated_summary", paper_and_summary.summary
                    )
                    relevant_papers.append(paper_and_summary)
            else:
                relevant_papers.append(Paper.from_entry(entry))
        return relevant_papers

    def get_past_relevant_papers(self) -> list[PaperAndSummary]:
        entries = self.reader.get_entries(read=True, tags=["is_relevant"])
        papers = []
        for ent in entries:
            generated_summary: str = dict(self.reader.get_tags(ent))[
                "generated_summary"
            ]
            paper = PaperAndSummary(Paper.from_entry(ent), generated_summary)
            papers.append(paper)
        return papers

    def count_unchecked_papers(self) -> int:
        self.reader.update_feeds()
        entry_counts = self.reader.get_entry_counts(read=False)
        return entry_counts.total if entry_counts.total else 0



from pathlib import Path
from typing import cast

import click
import sienna


def user_dir() -> Path:
    path = Path(click.get_app_dir("io.sotaro.akaibu"))
    path.mkdir(exist_ok=True, parents=True)
    return path


def get_endpoint() -> dict[str, str]:
    path = user_dir() / "endpoint.json"
    endpoint = cast(dict[str, str], sienna.load(path))
    return endpoint


def get_library_path(name: str | None) -> Path:
    name = "default" if not name else name
    path = user_dir() / f"{name}.sqlite"
    return path



from dataclasses import dataclass

from openai import OpenAI

from akaibu.paper import Paper, PaperAndSummary
from akaibu.utils import remove_reasoning_trace


@dataclass(frozen=True)
class PaperSummarizer:
    requirement: str
    model_name: str
    base_url: str
    key: str

    def summarize(self, paper: Paper) -> PaperAndSummary:
        client = OpenAI(base_url=self.base_url, api_key=self.key)
        response = client.chat.completions.create(
            model=self.model_name,
            messages=[
                {
                    "role": "user",
                    "content": f"""I have the following requirements and a research paper which meets the requirements. Summarize the paepr with a focus on the provided requirements in one sentence. Do not say that the paper matches the requirement, just produce a summary about the paper.
                    Requirements: {self.requirement}
                    Paper:
                      Title: {paper.title}
                      Abstract: {paper.abstract}""",
                },
            ],
        )
        res_cont = response.choices[0].message.content
        res_cont = res_cont if res_cont else ""
        res_cont = remove_reasoning_trace(res_cont).strip()

        return PaperAndSummary(paper=paper, summary=res_cont)


# @dataclass(frozen=True)
# class MultiSummarizer:
#     def summarize(self, papers: list[PaperAndSummary], model_name: str) -> str:
#         client = OpenAI(base_url=self.base_url, api_key=self.key)
#         paper_infos: str = "\n\n".join(
#             [f"Title: {p.paper.title}\nSummary: {p.summary}" for p in papers]
#         )
#         response = client.chat.completions.create(
#             model=model_name,
#             messages=[
#                 {
#                     "role": "user",
#                     "content": f"""Summarize the following list of papers (title and short summary) in one sentence like headline of a news letter. Use a few words to describe each paper.
#
#                     Papers:
#                       {paper_infos}""",
#                 },
#             ],
#         )
#         res_cont = response.choices[0].message.content
#         res_cont = res_cont if res_cont else ""
#         res_cont = re.sub(r"<think>.+<\/think>", "", res_cont, flags=re.DOTALL)
#         res_cont = res_cont if res_cont else ""
#         res_cont = res_cont.strip()
#
#         return res_cont



from typing import cast

import click
import sienna
from kensakun import Encoder, Engine
from rich.console import Console
from rich.table import Table

from akaibu import get_library_path, user_dir
from akaibu.checker import Checker
from akaibu.library import Library
from akaibu.paper import PaperAndSummary
from akaibu.summarizer import PaperSummarizer


@click.group()
def cli():
    pass


@cli.command(name="set-endpoint")
@click.argument("url")
@click.argument("key")
def set_endpoint(url: str, key: str):
    path = user_dir() / "endpoint.json"
    sienna.save({"url": url, "key": key}, path)


@cli.command(name="create-feed")
@click.argument("library_name")
@click.argument("url")
@click.argument("requirements")
def create_feed(library_name: str, url: str, requirements: str):
    path = user_dir() / "libraries.json"
    if path.exists():
        libraries: dict[str, str] = sienna.load(path)
    else:
        libraries: dict[str, str] = {}

    to_overwrite = False
    if library_name in libraries.keys():
        to_overwrite_str = input(
            f"{library_name} already exists. Do you want to overwrite it? (y/N)"
        )
        to_overwrite = to_overwrite_str == "y"
        if to_overwrite == "y":
            libraries[library_name] = requirements
        else:
            click.echo("Aborted.")
            return
    else:
        libraries[library_name] = requirements

    library_path = get_library_path(library_name)
    library = Library.load_from_path(library_path, remake=True)
    library.add_url(url)

    sienna.save(libraries, path)


@cli.command(name="digest")
@click.argument("n", default=5)
@click.argument("model_name")
@click.option(
    "library_name",
    "--library",
    "-l",
    default=None,
    help="Name of the target library (run `show-libraries` to check libraries you have). If not specified, all libraryes will be checked.",
)
@click.option(
    "in_markdown",
    "--in-markdown",
    "-m",
    is_flag=True,
    help="if given, the output will be formatetd as markdown, if not, will be a table",
)
@click.option(
    "no_sorting",
    "--no-sorting",
    "-ns",
    is_flag=True,
    help="If given, no semantic embeddings based sorting is applied, it just checks papers chronogically.",
)
def digest_n_documents(
    n: int,
    model_name: str,
    library_name: str | None,
    in_markdown: bool,
    no_sorting: bool,
):
    libraries = cast(dict[str, str], sienna.load(user_dir() / "libraries.json"))

    if library_name:
        if library_name not in libraries.keys():
            click.ClickException(
                f"{library_name} does not exist. Use `create-feed` command to make one."
            )
        library_names = [library_name]
    else:
        library_names = list(libraries.keys())

    # For kensakun
    if not no_sorting:
        encoder = Encoder.from_model_name(
            "Snowflake/snowflake-arctic-embed-m-v1.5",
            truncation_ratio=0.5,
        )
    else:
        encoder = None

    console = Console()
    for library_name in library_names:
        requirement: str = libraries[library_name]
        library_path = get_library_path(library_name)
        library = Library.load_from_path(library_path)

        if len(library.list_urls()) == 0:
            raise click.ClickException(
                "You have not set any arXiv urls. Use `set-feed` do set one."
            )

        library.reader.update_feeds()

        endpoint_path = user_dir() / "endpoint.json"
        if not endpoint_path.exists():
            raise click.ClickException(
                "You have not set endpoint information. Use `set-endpoint` command do set one."
            )

        endpoint: dict[str, str] = sienna.load(endpoint_path)
        base_url = endpoint["url"]
        key = endpoint["key"]

        # get summarizer
        summarizer = PaperSummarizer(requirement, model_name, base_url, key)

        # get checker
        checker = Checker(requirement, model_name, base_url, key)

        # Kensakun
        if not no_sorting:
            kensakun_engine = Engine.init(encoder, top_n=n)
        else:
            kensakun_engine = None

        # process papers
        papers = cast(
            list[PaperAndSummary],
            library.get_papers(n, checker, summarizer, kensakun=kensakun_engine),
        )

        # output
        if in_markdown:
            print("\n".join([p.to_markdown() for p in papers]))
        else:
            table = Table(
                title=library_name,
                show_header=True,
                header_style="bold magenta",
                show_lines=True,
            )
            table.add_column("Title")
            table.add_column("URL")
            table.add_column("Summary")

            for paper in papers:
                table.add_row(paper.paper.title, paper.paper.link, paper.summary)

            console.print(table)


@cli.command(name="remove-library")
@click.argument("name")
def remove_library(name: str):
    path = user_dir() / "libraries.json"
    if path.exists():
        libraries: dict[str, str] = sienna.load(path)
        libraries.pop(name)
        sienna.save(libraries, path)

        path = get_library_path(name)
        if path.exists() and path.is_file():
            path.unlink()
            (path.parent / f"{path.name}.search").unlink()


@cli.command(name="show-libraries")
def show_libraries():
    path = user_dir() / "libraries.json"
    if path.exists():
        libraries: dict[str, str] = sienna.load(path)

        console = Console()
        table = Table(show_header=True, header_style="bold magenta", show_lines=True)
        table.add_column("Name")
        table.add_column("URL")
        table.add_column("Requirement")
        table.add_column("# of Unchecked papers")

        for _name, _requirement in libraries.items():
            library = Library.load_from_path(get_library_path(_name), remake=False)
            url = list(library.reader.get_feeds())[0].url
            table.add_row(
                _name, url, _requirement, str(library.count_unchecked_papers())
            )

        console.print(table)


@cli.command(name="show-past-papers")
@click.argument("library_name")
@click.option(
    "in_markdown",
    "--in-markdown",
    "-m",
    is_flag=True,
)
def show_past_papers(library_name: str, in_markdown: bool):
    libraries = cast(dict[str, str], sienna.load(user_dir() / "libraries.json"))
    if library_name not in libraries.keys():
        click.ClickException(
            f"{library_name} does not exist. Use `create-feed` command to make one."
        )
    library_path = get_library_path(library_name)
    library = Library.load_from_path(library_path)

    papers = library.get_past_relevant_papers()

    if in_markdown:
        print(f"# {library_name}")
        print("\n".join([p.to_markdown() for p in papers]))
    else:
        console = Console()
        table = Table(show_header=True, header_style="bold magenta", show_lines=True)
        table.add_column("Title")
        table.add_column("URL")
        table.add_column("Summary")

        for paper in papers:
            table.add_row(paper.paper.title, paper.paper.link, paper.summary)

        console.print(table)
